import numpy as np
import anndata as ad
import time
import tracemalloc
import gc

from scmkl.train import train_model
from scmkl.test import predict, find_selected_groups
from scmkl.estimate_sigma import estimate_sigma
from scmkl.approx_kernels import calculate_z
from scmkl.multimodal_processing import multimodal_processing


def _check_adatas(adatas : list, check_uns : bool = False, 
                  check_obs : bool = False) -> None:
    '''
    Takes a list of AnnData objects and checks that all training and
    testing indices stored in adata.uns and cell labels stored in 
    adata.obs are the same. Ensures that all objects are of type 
    AnnData.
    Args:
        adatas - a list of AnnData objects where each element has the
            same array of indices for adata.uns['train_indices'] and 
            the same array of indices for adata.uns-'test_indices'].
            Additionally, adata.obs['labels'] must be the same for all
            AnnData objects.
        check_uns - a boolean value. If True, function will ensure that
            'train_indices' and 'test_indices' in adata.uns are the
            same. If False, indices will not be checked.
        check_obs - a boolean value. If True, function will ensure that
            adata.obs['labels'] in each AnnData object are the same.
    Returns:
        Returns None. Will throw an error if not all of the criteria
        listed above are met by adatas.
    '''
    for i, adata in enumerate(adatas):
        # Ensuring all elements are type AnnData
        if type(adata) != ad.AnnData: 
            raise TypeError(f'List object in position {i} is {type(adata)}, \
                            all elements in adatas must of type \
                            anndata.AnnData')
        
        # Ensuring all train/test indices are the same
        if check_uns:
            assert np.array_equal(adatas[0].uns['train_indices'],
                                    adata.uns['train_indices']), \
                "Train indices across AnnData objects in adatas do not \
                    match, ensure adata.uns['train_indices'] are the same \
                    for all elements"
                
            assert np.array_equal(adatas[0].uns['test_indices'], 
                                    adata.uns['test_indices']), \
                "Test indices across AnnData objects in adatas do not match, \
                    ensure adata.uns['test_indices'] are the same for all \
                    elements"

        # Ensuring all cell labels are the same 
        if check_obs:
            assert np.array_equal(adatas[0].obs['labels'], 
                                    adata.obs['labels']), \
                f"adata.obs['labels'] are different between AnnData objects"
    
    return None


def _eval_labels(cell_labels : np.ndarray, train_indices : np.ndarray, 
                  test_indices : np.ndarray, 
                  remove_labels : bool = False) -> np.ndarray:
    '''
    Takes an array of multiclass cell labels and returns a unique array 
    of cell labels to test for.
    Args:
        cell_labels - a numpy array of cell labels that coorespond to 
            an AnnData object.
        train_indices - a numpy array of indices for the training 
            samples in an AnnData object.
        test_indices - a numpy array of indices for the testing samples 
            in an AnnData object.
        remove_labels - If True, models will only be created for cell 
            labels in both the training and test data, if False, 
            models will be generated for all cell labels in the 
            training data.
    Returns:
        Returns a numpy array of unique cell labels to be iterated 
        through during one versus all experimental setups.
    '''
    train_uniq_labels = np.unique(cell_labels[train_indices])
    test_uniq_labels = np.unique(cell_labels[test_indices])

    # Ensuring that at least one cell type label between the two data
    #   are the same
    assert len(np.intersect1d(train_uniq_labels, test_uniq_labels)) > 0, \
        "There are no common labels between cells in the training and \
            testing samples"

    if np.array_equal(train_uniq_labels, test_uniq_labels):
        uniq_labels = train_uniq_labels
    else:
        # cells in the training but not the testing
        not_in_test = np.setdiff1d(train_uniq_labels, test_uniq_labels)
        # cells in the testing but not the training
        not_in_train = np.setdiff1d(test_uniq_labels, train_uniq_labels)    

        if len(not_in_test) > 0:
            print(f'WARNING: {not_in_test} cells are not in test data',
                flush = True)

        if len(not_in_train) > 0:
            print(f'WARNING: {not_in_train} cells are not in train data \
                  and will not be tested for',
                  flush = True)

        if remove_labels:
            uniq_labels = np.intersect1d(train_uniq_labels, test_uniq_labels)
        else:
            uniq_labels = train_uniq_labels

    return uniq_labels


def run(adata : ad.AnnData, alpha_list : np.ndarray, 
        metrics : list | None = None) -> dict:
    '''
    Takes a processed AnnData object and creates a model for each
    element in alpha_list. Return a dictionary with metrics and 
    predictions.
    Args:
        adata - an AnnData object with 'Z_train', 'Z_test', and 
            'group_dict' keys in adata.uns.
        alpha_list - a np.ndarray of alpha values to create models 
            using. Alpha refers to the penalty parameter in Group
            Lasso. Larger alphas force group weights to shrink towards
            0 while smaller alphas apply a lesser penalty to kernal 
            weights.
        metrics - a list of strings where each element is a type of 
            metric to be calculated. Options are ['AUROC', 'F1-Score',
            'Accuracy', 'Precision', 'Recall']. When set to None, all
            metrics are calculated.
    Returns:
        A dictionary with keys and values: 
            'Metrics' : a nested dictionary as [alpha][metric] = value
            'Selected_groups' : a dictionary as [alpha] = array of 
                groups with nonzero weights
            'Norms' : a dictionary as [alpha] = array of kernel weights
                for each group, order respective to 'Group_names'
            'Predictions' : a dictionary as [alpha] = predicted class
                respective to 'Observations' for that alpha
            'Observed' : an array of ground truth cell labels from the
                test set 
            'Test_indices' : indices of samples respective to adata 
                used in the training set
            'Group_names' : an array of group names respective to each
                array in 'Norms'
            'Model' : a dictionary where [alpha] = Celer Group Lasso
                object for that alpha
            'RAM_usage' : memory usage after training models for each 
                alpha
    '''
    if metrics == None:
        metrics = ['AUROC', 'F1-Score','Accuracy', 'Precision', 'Recall']

    # Initializing variables to capture metrics
    group_names = list(adata.uns['group_dict'].keys())
    predicted = {}
    group_norms = {}
    metric_dict = {}
    selected_pathways = {}
    train_time = {}
    models = {}

    D = adata.uns['D']

    # Generating models for each alpha and outputs
    for alpha in alpha_list:
        
        print(f'  Evaluating model. Alpha: {alpha}', flush = True)

        train_start = time.time()

        adata = train_model(adata, group_size= 2*D, alpha = alpha)
        predicted[alpha], metric_dict[alpha] = predict(adata, 
                                                            metrics = metrics)
        selected_pathways[alpha] = find_selected_groups(adata)

        kernel_weights = adata.uns['model'].coef_
        group_norms[alpha] = [
            np.linalg.norm(kernel_weights[i * 2 * D : (i + 1) * 2 * D - 1])
            for i in np.arange(len(group_names))
            ]
        
        models[alpha] = adata.uns['model']
        
        train_end = time.time()
        train_time[alpha] = train_end - train_start

    # Combining results into one object
    results = {}
    results['Metrics'] = metric_dict
    results['Selected_pathways'] = selected_pathways
    results['Norms'] = group_norms
    results['Predictions'] = predicted
    results['Observed'] = adata.obs['labels'].iloc[adata.uns['test_indices']]
    results['Test_indices'] = adata.uns['test_indices']
    results['Group_names']= group_names
    results['Models'] = models
    results['Train_time'] = train_time
    results['RAM_usage'] = f'{tracemalloc.get_traced_memory()[1] / 1e9} GB'

    return results



def one_v_all(adatas : list, names : list, alpha_list : np.ndarray, 
              tfidf : list, D : int, remove_labels = False) -> dict:
    '''
    This function takes a list of AnnData objects and creates a model
    for each cell label type.
    Args:
        adatas - a list of AnnData objects created by create_adata()
            where each AnnData is one modality and composed of both 
            training and testing samples. Requires that train_indices
            and test_indices are the same across all AnnDatas.
        names - a list of string variables that describe each modality
            respective to adatas for labeling.
        cell_labels - as scMKL is a binary classifier, there can only
            be two cell label types in the adata.obs['labels'] slot.
            Given this, np.ndarray is required to generate custom cell
            labels for each type present in the data. Array should 
            coorespond to samples in AnnData objects.
        remove_labels - If True, models will only be created for cell
            labels in both the training and test data, if False, models
            will be generated for all cell labels in the training data.
    Returns:
        Returns a dictionary of metrics, group weights, trained models
        for each label vs. the rest, predictions for each model, and a
        consensus prediction array for each sample in the train set.
    '''
    # Formatting checks ensuring all adata elements are 
    # AnnData objects and train/test indices are all the same
    _check_adatas(adatas, check_obs = True, check_uns = True)

    # Extracting train and test indices
    train_indices = adatas[0].uns['train_indices']
    test_indices = adatas[0].uns['test_indices']

    # Checking and capturing cell labels
    uniq_labels = _eval_labels(  cell_labels = adatas[0].obs['labels'], 
                                train_indices = train_indices,
                                 test_indices = test_indices,
                                remove_labels = remove_labels)


    # Calculating Z matrices, method depends on whether there are multiple 
    # adatas (modalities)
    if len(adatas) == 1:
        adata = estimate_sigma(adatas[0], n_features = 200)
        adata = calculate_z(adata, n_features = 5000)
    else:
        adata = multimodal_processing(adatas = adatas, 
                                            names = names, 
                                            tfidf = tfidf, 
                                            z_calculation = True)

    del adatas
    gc.collect()

    # Initializing for capturing model outputs
    results = {}

    # Capturing cell labels before overwriting
    cell_labels = np.array(adata.obs['labels'])

    for label in uniq_labels:
        print(f"Comparing {label} to other types", flush = True)
        cur_labels = cell_labels.copy()
        cur_labels[cell_labels != label] = 'other'
        
        # Replacing cell labels for current cell type vs rest
        adata.obs['labels'] = cur_labels

        # Running scMKL
        results[label] = run(adata, alpha_list)


    return results


import numpy as np
import gc
import tracemalloc

from scmkl.tfidf import tfidf_normalize
from scmkl.estimate_sigma import estimate_sigma
from scmkl.approx_kernels import calculate_z
from scmkl.train import train_model
from scmkl.test import calculate_auroc, find_selected_groups
from scmkl.multimodal_processing import multimodal_processing


def optimize_alpha(adata, group_size, tfidf = False, alpha_array = np.round(np.linspace(1.9,0.1, 10),2), k = 4):
    '''
    Iteratively train a grouplasso model and update alpha to find the parameter yielding the desired sparsity.
    This function is meant to find a good starting point for your model, and the alpha may need further fine tuning.
    Input:
        adata- Anndata object with Z_train and Z_test calculated
        group_size- Argument describing how the features are grouped. 
            From Celer documentation:
            "groupsint | list of ints | list of lists of ints.
                Partition of features used in the penalty on w. 
                    If an int is passed, groups are contiguous blocks of features, of size groups. 
                    If a list of ints is passed, groups are assumed to be contiguous, group number g being of size groups[g]. 
                    If a list of lists of ints is passed, groups[g] contains the feature indices of the group number g."
            If 1, model will behave identically to Lasso Regression.
        tfidf- Boolean value to determine if TFIDF normalization should be run at each fold. True means that it will be performed.
        starting_alpha- The alpha value to start the search at.
        alpha_array- Numpy array of all alpha values to be tested
        k- number of folds to perform cross validation over
            
    Output:
        sparsity_dict- Dictionary with tested alpha as keys and the number of selected pathways as the values
        alpha- The alpha value yielding the number of selected groups closest to the target.
    '''

    assert isinstance(k, int) and k > 0, 'Must be a positive integer number of folds'

    import warnings 
    warnings.filterwarnings('ignore')

    y = adata.obs['labels'].iloc[adata.uns['train_indices']].to_numpy()
    
    # Splits the labels evenly between folds
    positive_indices = np.where(y == np.unique(y)[0])[0]
    negative_indices = np.setdiff1d(np.arange(len(y)), positive_indices)

    positive_annotations = np.arange(len(positive_indices)) % k
    negative_annotations = np.arange(len(negative_indices)) % k

    auc_array = np.zeros((len(alpha_array), k))

    gc.collect()

    for fold in np.arange(k):
        
        print(f'Fold {fold + 1}:\n Memory Usage: {[mem / 1e9 for mem in tracemalloc.get_traced_memory()]} GB')

        cv_adata = adata[adata.uns['train_indices'],:]

        # Create CV train/test indices
        fold_train = np.concatenate((positive_indices[np.where(positive_annotations != fold)[0]], negative_indices[np.where(negative_annotations != fold)[0]]))
        fold_test = np.concatenate((positive_indices[np.where(positive_annotations == fold)[0]], negative_indices[np.where(negative_annotations == fold)[0]]))

        cv_adata.uns['train_indices'] = fold_train
        cv_adata.uns['test_indices'] = fold_test

        if tfidf:
            cv_adata = tfidf_normalize(cv_adata, binarize= True)

        cv_adata = estimate_sigma(cv_adata, n_features = 200)
        cv_adata = calculate_z(cv_adata, n_features= 5000)

        gc.collect()

        for i, alpha in enumerate(alpha_array):

            cv_adata = train_model(cv_adata, group_size, alpha = alpha)
            auc_array[i, fold] = calculate_auroc(cv_adata)

            gc.collect()

        del cv_adata
        gc.collect()
        
    # Take AUROC mean across the k folds to find alpha yielding highest AUROC
    alpha_star = alpha_array[np.argmax(np.mean(auc_array, axis = 1))]
    gc.collect()
    

    return alpha_star


def multimodal_optimize_alpha(adatas : list, group_size = 1, tfidf = [False, False], alpha_array = np.round(np.linspace(1.9,0.1, 10),2), k = 4):
    '''
    Iteratively train a grouplasso model and update alpha to find the parameter yielding the desired sparsity.
    This function is meant to find a good starting point for your model, and the alpha may need further fine tuning.
    Input:
        adatas- a list of AnnData objects where each object is one modality and Z_train and Z_test are calculated
        group_size- Argument describing how the features are grouped. 
            From Celer documentation:
            "groupsint | list of ints | list of lists of ints.
                Partition of features used in the penalty on w. 
                    If an int is passed, groups are contiguous blocks of features, of size groups. 
                    If a list of ints is passed, groups are assumed to be contiguous, group number g being of size groups[g]. 
                    If a list of lists of ints is passed, groups[g] contains the feature indices of the group number g."
            If 1, model will behave identically to Lasso Regression.
        tifidf_list- a boolean mask where tfidf_list[0] and tfidf_list[1] are respective to adata1 and adata2
            If True, tfidf normalization will be applied to the respective adata during cross validation
        starting_alpha- The alpha value to start the search at.
        alpha_array- Numpy array of all alpha values to be tested
        k- number of folds to perform cross validation over
            
    Output:
        sparsity_dict- Dictionary with tested alpha as keys and the number of selected pathways as the values
        alpha- The alpha value yielding the number of selected groups closest to the target.
    '''

    assert isinstance(k, int) and k > 0, 'Must be a positive integer number of folds'

    import warnings 
    warnings.filterwarnings('ignore')

    y = adatas[0].obs['labels'].iloc[adatas[0].uns['train_indices']].to_numpy()
    
    # Splits the labels evenly between folds
    positive_indices = np.where(y == np.unique(y)[0])[0]
    negative_indices = np.setdiff1d(np.arange(len(y)), positive_indices)

    positive_annotations = np.arange(len(positive_indices)) % k
    negative_annotations = np.arange(len(negative_indices)) % k

    auc_array = np.zeros((len(alpha_array), k))

    cv_adatas = []

    for adata in adatas:
        cv_adatas.append(adata[adata.uns['train_indices'],:].copy())

    del adatas
    gc.collect()

    for fold in np.arange(k):
        
        print(f'Fold {fold + 1}:\n Memory Usage: {[mem / 1e9 for mem in tracemalloc.get_traced_memory()]} GB')

        fold_train = np.concatenate((positive_indices[np.where(positive_annotations != fold)[0]], negative_indices[np.where(negative_annotations != fold)[0]]))
        fold_test = np.concatenate((positive_indices[np.where(positive_annotations == fold)[0]], negative_indices[np.where(negative_annotations == fold)[0]]))

        for i in range(len(cv_adatas)):
            cv_adatas[i].uns['train_indices'] = fold_train
            cv_adatas[i].uns['test_indices'] = fold_test

        # Creating dummy names for cv. 
        # #Necessary for interpretability but not for AUROC cv
        dummy_names = [f'mod{i}' for i in range(len(cv_adatas))]

        # Calculate the Z's for each modality independently
        fold_cv_adata = multimodal_processing(adatas = cv_adatas, names = dummy_names, tfidf = tfidf, z_calculation = True)
        fold_cv_adata.uns['seed_obj'] = cv_adatas[0].uns['seed_obj']

        gc.collect()

        for j, alpha in enumerate(alpha_array):

            fold_cv_adata = train_model(fold_cv_adata, group_size, alpha = alpha)

            auc_array[j, fold] = calculate_auroc(fold_cv_adata)

        del fold_cv_adata
        gc.collect()

    # Take AUROC mean across the k folds and select the alpha resulting in highest AUROC
    alpha_star = alpha_array[np.argmax(np.mean(auc_array, axis = 1))]
    del cv_adatas
    gc.collect()
    
    return alpha_star


def optimize_sparsity(adata, group_size, starting_alpha = 1.9, increment = 0.2, target = 1, n_iter = 10):
    '''
    Iteratively train a grouplasso model and update alpha to find the parameter yielding the desired sparsity.
    This function is meant to find a good starting point for your model, and the alpha may need further fine tuning.
    Input:
        adata- Anndata object with Z_train and Z_test calculated
        group_size- Argument describing how the features are grouped. 
            From Celer documentation:
            "groupsint | list of ints | list of lists of ints.
                Partition of features used in the penalty on w. 
                    If an int is passed, groups are contiguous blocks of features, of size groups. 
                    If a list of ints is passed, groups are assumed to be contiguous, group number g being of size groups[g]. 
                    If a list of lists of ints is passed, groups[g] contains the feature indices of the group number g."
            If 1, model will behave identically to Lasso Regression.
        starting_alpha- The alpha value to start the search at.
        increment- amount to adjust alpha by between iterations
        target- The desired number of groups selected by the model.
        n_iter- The maximum number of iterations to run
            
    Output:
        sparsity_dict- Dictionary with tested alpha as keys and the number of selected pathways as the values
        alpha- The alpha value yielding the number of selected groups closest to the target.
    '''
    assert increment > 0 and increment < starting_alpha, 'Choose a positive increment less than alpha'
    assert target > 0 and isinstance(target, int), 'Choose an integer target number of groups that is greater than 0'
    assert n_iter > 0 and isinstance(n_iter, int), 'Choose an integer number of iterations that is greater than 0'

    sparsity_dict = {}
    alpha = starting_alpha

    for _ in np.arange(n_iter):
        adata = train_model(adata, group_size, alpha)
        num_selected = len(find_selected_groups(adata))

        sparsity_dict[np.round(alpha,4)] = num_selected

        if num_selected < target:
            #Decreasing alpha will increase the number of selected pathways
            if alpha - increment in sparsity_dict.keys():
                # Make increment smaller so the model can't go back and forth between alpha values
                increment /= 2
            alpha = np.max([alpha - increment, 1e-1]) #Ensures that alpha will never be negative
        elif num_selected > target:
            if alpha + increment in sparsity_dict.keys():
                increment /= 2
            alpha += increment
        elif num_selected == target:
            break

    # Find the alpha that minimizes the difference between target and observed number of selected groups
    optimal_alpha = list(sparsity_dict.keys())[np.argmin([np.abs(selected - target) for selected in sparsity_dict.values()])]
    return sparsity_dict, optimal_alpha


import numpy as np
import anndata as ad
import gc

from scmkl.tfidf import tfidf_normalize
from scmkl.estimate_sigma import estimate_sigma
from scmkl.approx_kernels import calculate_z


def combine_modalities(adatas : list,
                       names : list,
                       combination = 'concatenate'):
    '''
    Combines data sets for multimodal classification.  Combined group names are assay+group_name
    Input:
            adatas: a list of AnnData objects where each object is a different modality
            names: a list of strings names for each modality repective to each object in adatas
            combination: How to combine the matrices, either sum or concatenate
    Output:
            combined_adata: Adata object with the combined Z matrices and annotations.  Annotations must match
    '''
    # assert assay_1_adata.shape[0] == assay_2_adata.shape[0], 'Cannot combine data with different number of cells.'
    assert len(np.unique(names)) == len(names), 'Assay names must be distinct'
    assert combination.lower() in ['sum', 'concatenate']
    assert all([(('Z_train' in adata.uns.keys()) & ('Z_test' in adata.uns.keys())) for adata in adatas]), 'Z not calculated'

    combined_adata = ad.concat(adatas, uns_merge = 'same', axis = 1, label = 'labels')

    assert 'train_indices' in combined_adata.uns.keys(), 'Different train test splits between AnnData objects'

    # Conserving labels from adatas
    combined_adata.obs = adatas[0].obs

    # Creating a single dictionary with all of the groups across modalities 
    group_dict = {}
    for name, adata in zip(names, adatas):
        for group_name, features in adata.uns['group_dict'].items():
            group_dict[f'{name}-{group_name}'] = features

    if combination == 'concatenate':
        combined_adata.uns['Z_train'] = np.hstack([adata.uns['Z_train'] for adata in adatas])
        combined_adata.uns['Z_test'] = np.hstack([adata.uns['Z_test'] for adata in adatas])


    elif combination == 'sum':

        #Check that the dimensions of all Z's are the same before concatentation
        dims = [adata.uns['Z_train'].shape for adata in adatas]
        dims = all([dim == dims[0] for dim in dims])
        assert dims, 'Cannot sum Z matrices with different dimensions'
        
        combined_adata.uns['Z_train'] = np.sum([adata.uns['Z_train'] for adata in adatas], axis = 0)
        combined_adata.uns['Z_test'] = np.sum([adata.uns['Z_test'] for adata in adatas], axis = 0)


    combined_adata.uns['group_dict'] = group_dict

    if 'seed_obj' in adatas[0].uns_keys():
        combined_adata.uns['seed_obj'] = adatas[0].uns['seed_obj']
    else:
        print('No random seed present in Adata, it is recommended for reproducibility.')

    del adatas
    gc.collect()

    return combined_adata


def multimodal_processing(adatas : list, names : list, tfidf: list, z_calculation = False):
    '''
    Function to remove rows from both modalities when there is no signal present in at least 1.

    Input:
        adatas: a list of AnnData objects where each object is a different modality
        names: a list of string names for each modality repective to each object in adatas
        tfidf- list of boolean values whether to tfidf the respective matrices
        z_calculation- Boolean value whether to calculate sigma and Z on the adata before combining
                        Allows for individual kernel functions for each adata
    Output:
        adata- Concatenated adata objects with empty rows removed and matching order
    '''

    import warnings 
    warnings.filterwarnings('ignore')

    assert all([adata.shape[0] for adata in adatas]), 'Different number of cells present in each object'
    assert np.all([np.array_equal(adatas[0].uns['train_indices'], adatas[i].uns['train_indices']) for i in range(1, len(adatas))]), 'Different train indices'
    assert np.all([np.array_equal(adatas[0].uns['test_indices'], adatas[i].uns['test_indices']) for i in range(1, len(adatas))]), 'Different test indices'

    # Getting a list of the rows that are not empty across all of the input modalities
    # Creates a boolean array for each modality of cells with non-empty rows
    non_empty_rows = [np.array((np.sum(adata.X, axis = 1) > 0)).ravel() for adata in adatas]

    # Compares all nonempty boolean arrays and returns a 1d array where sample feature sums
    #   across all modalities are more than 0
    non_empty_rows = np.logical_and(*non_empty_rows).squeeze()

    # Initializing final train test split array
    train_test = np.repeat('train', adatas[0].shape[0])
    train_test[adatas[0].uns['test_indices']] = 'test'

    # Capturing train test split with empty rows filtered out
    train_test = train_test[non_empty_rows]
    train_indices = np.where(train_test == 'train')[0]
    test_indices = np.where(train_test == 'test')[0]

    # Adding train test split arrays to AnnData objects and filtering out empty samples
    for i, adata in enumerate(adatas):
        adata.uns['train_indices'] = train_indices
        adata.uns['test_indices'] = test_indices
        adata = adata[non_empty_rows, :]
        # tfidf normalizing if corresponding element in tfidf is True
        if tfidf[i]:
            adata = tfidf_normalize(adata)

        if z_calculation:
            # AnnData update must be pointing at the object in list
            print(f'Estimating Sigma for {names[i]}', flush = True)
            adatas[i] = estimate_sigma(adata, n_features= 200)
            print(f'Calculating Z for {names[i]}', flush = True)
            adatas[i] = calculate_z(adata, n_features = 5000)

    if 'labels' in adatas[0].obs:
        all_labels = [adata.obs['labels'] for adata in adatas]
        # Ensuring cell labels for each AnnData object are the same
        for i in range(1, len(all_labels)):
            assert np.all(all_labels[0] == all_labels[i]), f'Cell labels between AnnData object in position 0 and position {i} in adatas do not match'

    adata = combine_modalities(adatas = adatas,
                               names = names,
                                combination = 'concatenate')

    del adatas
    gc.collect()

    return adata    




import numpy as np
import scipy


def _tfidf(X, mode = 'filter'):
    '''
    Function to use Term Frequency Inverse Document Frequency filtering for atac data to find meaningful features. 
    If input is pandas data frame or scipy sparse array, it will be converted to a numpy array.
    Input:
            x- Data matrix of cell x feature.  Must be a Numpy array or Scipy sparse array.
            mode- Argument to determine what to return.  Must be filter or normalize
    Output:
            TFIDF- Output depends on given 'mode' parameter
                'filter'- returns which column sums are non 0 i.e. which features are significant
                'normalize'- returns TFIDF filtered data matrix of the same dimensions as x. Returns as scipy sparse matrix
    '''

    assert mode in ['filter', 'normalize'], 'mode must be "filter" or "normalize".'
    
    if scipy.sparse.issparse(X):
        tf = scipy.sparse.csc_array(X)
        doc_freq = np.array(np.sum(X > 0, axis=0)).flatten()
    else:
        tf = X
        doc_freq = np.sum(X > 0, axis=0)

    idf = np.log1p((1 + X.shape[0]) / (1 + doc_freq))
    tfidf = tf * idf

    if mode == 'normalize':
        if scipy.sparse.issparse(tfidf):
            tfidf = scipy.sparse.csc_matrix(tfidf)
        return tfidf
    elif mode == 'filter':
        significant_features = np.where(np.sum(tfidf, axis=0) > 0)[0]
        return significant_features
        

def tfidf_normalize(adata, binarize = False):

    '''
    Function to TF IDF normalize the data in an adata object
    If train/test indices are included in the object, it will calculate the normalization separately for the training and testing data
        Otherwise it will calculate it on the entire dataset
    If any rows are entirely 0, that row and its metadata will be removed from the object

    Input:
        adata- adata object with data in adata.X to be normalized
            Can have train/test indices included or not
        binarize- Boolean option to binarize the data
    Output:
        adata- adata object with same attributes as before, but the TF IDF normalized matrix in place of adata.X
                    Will now have the train data stacked on test data, and the indices will be adjusted accordingly
    '''

    X = adata.X.copy()
    row_sums = np.sum(X, axis = 1)
    assert np.all(row_sums > 0), 'TFIDF requires all row sums be positive'

    if binarize:
        X[X>0] = 1

    if 'train_indices' in adata.uns_keys():

        train_indices = adata.uns['train_indices'].copy()
        test_indices = adata.uns['test_indices'].copy()

        # Calculate the train TFIDF matrix on just the training data so it is not biased by testing data
        tfidf_train = _tfidf(X[train_indices,:], mode = 'normalize')

        # Calculate the test TFIDF by calculating it on the train and test data and index the test data
        tfidf_test = _tfidf(X, mode = 'normalize')[test_indices,:]

        # Impossible to add rows back to original location so we need to stack the matrices to maintain train/test
        if scipy.sparse.issparse(X):
            tfidf_norm = scipy.sparse.vstack((tfidf_train, tfidf_test))
        else:
            tfidf_norm = np.vstack((tfidf_train, tfidf_test))

        ## I'm not sure why this reassignment is necessary, but without, the values will be saved as 0s in adata
        adata.uns['train_indices'] = train_indices
        adata.uns['test_indices'] = test_indices

        combined_indices = np.concatenate((train_indices, test_indices))

        # Anndata indexes by "rownames" not position so we need to rename the rows to properly index
        adata_index = adata.obs_names[combined_indices].astype(int)
        tfidf_norm = tfidf_norm[np.argsort(adata_index),:]

    else:

        tfidf_norm = _tfidf(X, mode = 'normalize')

    adata.X = tfidf_norm.copy()

    return adata


import numpy as np
import scipy
import anndata as ad


def filter_features(X, feature_names, group_dict):
    '''
    Function to remove unused features from X matrix.  Any features not included in group_dict will be removed from the matrix.
    Also puts the features in the same relative order (of included features)
    Input:
            X- Data array. Can be Numpy array or Scipy Sparse Array
            feature_names- Numpy array of corresponding feature names
            group_dict- Dictionary containing feature grouping information.
                        Example: {geneset: np.array(gene_1, gene_2, ..., gene_n)}
    Output:
            X- Data array containing data only for features in the group_dict
            feature_names- Numpy array of corresponding feature names from group_dict
    '''
    assert X.shape[1] == len(feature_names), 'Given features do not correspond with features in X'    

    group_features = set()
    feature_set = set(feature_names)

    # Store all objects in dictionary in set
    for group in group_dict.keys():
        group_features.update(set(group_dict[group]))

        # Finds intersection between group features and features in data
        # Converts to numpy array and sorts to preserve order of feature names
        group_dict[group] = np.sort(np.array(list(feature_set.intersection(set(group_dict[group])))))

    # Find location of desired features in whole feature set
    group_feature_indices = np.where(np.in1d(feature_names, np.array(list(group_features)), assume_unique = True))[0]

    # Subset only the desired features and their data
    X = X[:,group_feature_indices]
    feature_names = np.array(list(feature_names))[group_feature_indices]

    return X, feature_names, group_dict


def train_test_split(y, train_indices = None, seed_obj = np.random.default_rng(100), train_ratio = 0.8):
    '''
    Function to calculate training and testing indices for given dataset. If train indices are given, it will calculate the test indices.
        If train_indices == None, then it calculates both indices, preserving the ratio of each label in y
    Input:
            y- Numpy array of cell labels. Can have any number of classes for this function.
            train_indices- Optional array of pre-determined training indices
            seed_obj- Numpy random state used for random processes. Can be specified for reproducubility or set by default.
            train_ratio- decimal value ratio of features in training:testing sets
    Output:
            train_indices- Array of indices of training cells
            test_indices- Array of indices of testing cells
    '''

    # If train indices aren't provided
    if train_indices is None:

        unique_labels = np.unique(y)
        train_indices = []

        for label in unique_labels:

            # Find index of each unique label
            label_indices = np.where(y == label)[0]

            # Sample these indices according to train ratio
            train_label_indices = seed_obj.choice(label_indices, int(len(label_indices) * train_ratio), replace = False)
            train_indices.extend(train_label_indices)
    else:
        assert len(train_indices) <= len(y), 'More train indices than there are samples'

    train_indices = np.array(train_indices)

    # Test indices are the indices not in the train_indices
    test_indices = np.setdiff1d(np.arange(len(y)), train_indices, assume_unique = True)

    return train_indices, test_indices

def sparse_var(X, axis = None):

    '''
    Function to calculate variance on a scipy sparse matrix.
    Input:
        X- A scipy sparse or numpy array
        axis- Determines which axis variance is calculated on. Same usage as Numpy
            axis = 0 => column variances
            axis = 1 => row variances
            axis = None => total variance (calculated on all data)
    Output:
        var- Variance values calculated over the given axis
    '''

    # E[X^2] - E[X]^2
    if scipy.sparse.issparse(X):
        var = np.array((X.power(2).mean(axis = axis)) - np.square(X.mean(axis = axis)))
    else:
        var = np.var(X, axis = axis)
    return var.ravel()

def process_data(X_train, X_test = None, data_type = 'counts', return_dense = True):


    '''
    Function to preprocess data matrix according to type of data (counts- e.g. rna, or binary- atac)
    Will process test data according to parameters calculated from test data

    Input:
        X_train- A scipy sparse or numpy array
        X_train- A scipy sparse or numpy array
        data_type- 'counts' or 'binary'.  Determines what preprocessing is applied to the data. 
            Log transforms and standard scales counts data
            TFIDF filters ATAC data to remove uninformative columns
    Output:
        X_train, X_test- Numpy arrays with the process train/test data respectively.
    '''

    if data_type not in ['counts', 'binary']:
        print('Data will not be normalized for gene expression data')
        print('Columns with zero summed columns will not be removed')
        print('To change this behavior, set data_type to counts or binary')

    if X_test is None:
            X_test = X_train[:1,:] # Creates dummy matrix to for the sake of calculation without increasing computational time
            orig_test = None
    else:
        orig_test = 'given'

    # Remove features that have no variance in the training data (will be uniformative)
    var = sparse_var(X_train, axis = 0)
    variable_features = np.where(var > 1e-5)[0]

    X_train = X_train[:,variable_features]
    X_test = X_test[:, variable_features]

    #Data processing according to data type
    if data_type.lower() == 'counts':

        if scipy.sparse.issparse(X_train):
            X_train = X_train.log1p()
            X_test = X_test.log1p()
        else:
            X_train = np.log1p(X_train)
            X_test = np.log1p(X_test)
            
        #Center and scale count data
        train_means = np.mean(X_train, 0)
        train_sds = np.sqrt(var[variable_features])

        # Perform transformation on test data according to parameters of the training data
        X_train = (X_train - train_means) / train_sds
        X_test = (X_test - train_means) / train_sds

    if return_dense and scipy.sparse.issparse(X_train):
        X_train = X_train.toarray()
        X_test = X_test.toarray()

    if orig_test == None:
        return X_train
    else:
        return X_train, X_test


def create_adata(X, feature_names: np.ndarray, cell_labels: np.ndarray, group_dict: dict, data_type: str, split_data = None, D = 100, 
                 remove_features = True, distance_metric = 'euclidean', kernel_type = 'Gaussian', random_state = 1):
    
    '''
    Function to create an AnnData object to carry all relevant information going forward

    Input:
        X- A data matrix of cells by features can be a numpy array, scipy sparse array or pandas dataframe (sparse array recommended for large datasets)
        feature_names- A numpy array of feature names corresponding with the features in X
        cell_labels- A numpy array of cell phenotypes corresponding with the cells in X.  Must be binary
        group_dict- Dictionary containing feature grouping information.
                        Example: {geneset: np.array(gene_1, gene_2, ..., gene_n)}
        data_type- 'counts' or 'binary'.  Determines what preprocessing is applied to the data. 
            Log transforms and standard scales counts data
            TFIDF filters binary data
        split_data- Either numpy array of precalculated train/test split for the cells -or-
                            None.  If None, the train test split will be calculated with balanced classes.
        D- Number of Random Fourier Features used to calculate Z. Should be a positive integer.
                Higher values of D will increase classification accuracy at the cost of computation time
        remove_features- Bool whether to filter the features from the dataset.
                Will remove features from X and feature_names not in group_dict and remove features from groupings not in feature_names
        distance_metric- The pairwise distance metric used to estimate sigma.  
                Must be one of the options used in scipy.spatial.distance.cdist
        kernel_type- The approximated kernel function used to calculate Zs
                Must be one of Gaussian, Laplacian, or Cauchy
        random_state- Integer random_state used to set the seed for reproducibilty.
    Output:
        AnnData object with: 
            Data equal to the values in X- accessible with adata.X
            Variable names equal to the values in feature_names- accessible with adata.var_names
            Cell phenotypes equal to the values in cell_labels- accessible with adata.obs['labels']
            Train/test split either as given or calculated in this function- accessible with adata.uns['train_indices'] and adata.uns['test_indices'] respectively
            Grouping information equal to given group_dict- accessible with adata.uns['group_dict']
            seed_obj with seed equal to 100 * random_state- accessible with adata.uns['seed_obj']
            D- accessible with adata.uns['D']
            Type of data to determine preprocessing steps- accessible with adata.uns['data_type']
            Distance metric as given- accessible with adata.uns['distance_metric']
            Kernel function as given- accessible with adata.uns['kernel_type']

    '''

    assert X.shape[0] == len(cell_labels), 'Different number of cells than labels'
    assert X.shape[1] == len(feature_names), 'Different number of features in X than feature names'
    assert len(np.unique(cell_labels)) == 2, 'cell_labels must contain 2 classes'
    assert isinstance(D, int) and D > 0, 'D must be a positive integer'
    assert kernel_type.lower() in ['gaussian', 'laplacian', 'cauchy'], 'Given kernel type not implemented. Gaussian, Laplacian, and Cauchy are the acceptable types.'

    if remove_features:
        X, feature_names, group_dict = filter_features(X, feature_names, group_dict)

    # Create adata object and add column names
    adata = ad.AnnData(X)
    adata.var_names = feature_names

    # Add metadata to adata object
    adata.obs['labels'] = cell_labels
    adata.uns['group_dict'] = group_dict
    adata.uns['seed_obj'] = np.random.default_rng(100 * random_state)
    adata.uns['data_type'] = data_type
    adata.uns['D'] = D
    adata.uns['kernel_type'] = kernel_type
    adata.uns['distance_metric'] = distance_metric

    if split_data == None:
        train_indices, test_indices = train_test_split(cell_labels, seed_obj = adata.uns['seed_obj'])
    else:
        train_indices = np.where(split_data == 'train')[0]
        test_indices = np.where(split_data == 'test')[0]

    adata.uns['train_indices'] = train_indices
    adata.uns['test_indices'] = test_indices

    return adata






import numpy as np
import sklearn


def predict(adata, metrics = None):
    '''
    Function to return predicted labels and calculate any of AUROC, Accuracy, F1 Score, Precision, Recall for a classification. 
    Input:  
            adata- adata object with trained model and Z matrices in uns
            metrics- Which metrics to calculate on the predicted values

    Output:
            Values predicted by the model
            Dictionary containing AUROC, Accuracy, F1 Score, Precision, and/or Recall depending on metrics argument

    '''
    y_test = adata.obs['labels'].iloc[adata.uns['test_indices']].to_numpy()
    X_test = adata.uns['Z_test']
    assert X_test.shape[0] == len(y_test), 'X and y must have the same number of samples'
    assert all([metric in ['AUROC', 'Accuracy', 'F1-Score', 'Precision', 'Recall'] for metric in metrics]), 'Unknown metric provided.  Must be one or more of AUROC, Accuracy, F1-Score, Precision, Recall'

    # Sigmoid function to force probabilities into [0,1]
    probabilities = 1 / (1 + np.exp(-adata.uns['model'].predict(X_test)))

    # Group Lasso requires 'continous' y values need to re-descritize it
    y = np.zeros((len(y_test)))
    y[y_test == np.unique(y_test)[0]] = 1

    metric_dict = {}

    #Convert numerical probabilities into binary phenotype
    y_pred = np.array(np.repeat(np.unique(y_test)[1], len(y_test)), dtype = 'object')
    y_pred[np.round(probabilities,0).astype(int) == 1] = np.unique(y_test)[0]

    if metrics == None:
        return y_pred
    
    # Calculate and save metrics given in metrics
    if 'AUROC' in metrics:
        fpr, tpr, _ = sklearn.metrics.roc_curve(y, probabilities)
        metric_dict['AUROC'] = sklearn.metrics.auc(fpr, tpr)
    if 'Accuracy' in metrics:
        metric_dict['Accuracy'] = np.mean(y_test == y_pred)
    if 'F1-Score' in metrics:
        metric_dict['F1-Score'] = sklearn.metrics.f1_score(y_test, y_pred, pos_label = np.unique(y_test)[0])
    if 'Precision' in metrics:
        metric_dict['Precision'] = sklearn.metrics.precision_score(y_test, y_pred, pos_label = np.unique(y_test)[0])
    if 'Recall' in metrics:
        metric_dict['Recall'] = sklearn.metrics.recall_score(y_test, y_pred, pos_label = np.unique(y_test)[0])

    return y_pred, metric_dict


def find_selected_groups(adata) -> np.ndarray:

    '''
    Function to find feature groups selected by the model during training.  If feature weight assigned by the model is non-0, then the group containing that feature is selected.
    Inputs:
        model- A trained celer.GroupLasso model.
        group_names- An iterable object containing the group_names in the same order as the feature groupings from Data array.
    Outpus:
        Numpy array containing the names of the groups selected by the model.
    '''

    selected_groups = []
    coefficients = adata.uns['model'].coef_
    group_size = adata.uns['model'].get_params()['groups']
    group_names = np.array(list(adata.uns['group_dict'].keys()))

    # Loop over the model weights associated with each group and calculate the L2 norm.
    for i, group in enumerate(group_names):
        if not isinstance(group_size, (list, set, np.ndarray, tuple)):
            group_norm = np.linalg.norm(coefficients[np.arange(i * group_size, (i+1) * group_size - 1)])
        else: 
            group_norm = np.linalg.norm(coefficients[group_size[i]])

        # Only include the group if the model weights are > 0 
        if group_norm != 0:
            selected_groups.append(group)

    return np.array(selected_groups)


def calculate_auroc(adata)-> float:
    '''
    Function to calculate the AUROC for a classification. 
    Designed as a helper function.  Recommended to use Predict() for model evaluation.
    Input:  
            adata- adata object with trained model and Z matrices in uns
    Output:
            Calculated AUROC value
    '''

    y_test = adata.obs['labels'].iloc[adata.uns['test_indices']].to_numpy()
    X_test = adata.uns['Z_test']

    y_test = y_test.ravel()
    assert X_test.shape[0] == len(y_test), f'X has {X_test.shape[0]} samples and y has {len(y_test)} samples.'

    # Sigmoid function to force probabilities into [0,1]
    probabilities = 1 / (1 + np.exp(-adata.uns['model'].predict(X_test)))
    # Group Lasso requires 'continous' y values need to re-descritize it

    y = np.zeros((len(y_test)))
    y[y_test == np.unique(y_test)[0]] = 1
    fpr, tpr, _ = sklearn.metrics.roc_curve(y, probabilities)
    auc = sklearn.metrics.auc(fpr, tpr)
    
    return(auc)


def extract_kernel_weights():
    '''
    '''
    pass


import numpy as np
import scipy

from scmkl.data_processing import process_data


def estimate_sigma(adata, n_features = 5000):
    '''
    Function to calculate approximate kernels widths to inform distribution for project of Fourier Features. Calculates one sigma per group of features
    Input:
            adata- Adata obj as created by `Create_Adata`
            n_features- Number of random features to include when estimating sigma.  Will be scaled for the whole pathway set according to a heuristic. Used for scalability
    Output:
            adata object with sigma values.  Sigmas accessible by adata.uns['sigma']

    '''
 
    sigma_list = []

    # Loop over every group in group_dict
    for group_features in adata.uns['group_dict'].values():

        # Select only features within that group and downsample for scalability
        num_group_features = len(group_features)
        group_features = adata.uns['seed_obj'].choice(np.array(list(group_features)), min([n_features, num_group_features]), replace = False) 

        # Use on the train data to estimate sigma
        X_train = adata[adata.uns['train_indices'], group_features].X
        X_train = process_data(X_train = X_train, data_type = adata.uns['data_type'], return_dense = True)
        
        # Sample cells because distance calculation are costly and can be approximated
        distance_indices = adata.uns['seed_obj'].choice(np.arange(X_train.shape[0]), np.min((2000, X_train.shape[0])))

        # Calculate Distance Matrix with specified metric
        sigma = np.mean(scipy.spatial.distance.cdist(X_train[distance_indices,:], X_train[distance_indices,:], adata.uns['distance_metric']))

        # sigma = 0 is numerically unusable in later steps
        # Using such a small sigma will result in wide distribution, and typically a non-predictive Z
        if sigma == 0:
            sigma += 1e-5

        if n_features < num_group_features:
            sigma = sigma * num_group_features / n_features # Heuristic we calculated to account for fewer features used in distance calculation

        sigma_list.append(sigma)
    
    adata.uns['sigma'] = np.array(sigma_list)
        
    return adata


import numpy as np

from scmkl.data_processing import process_data


def calculate_z(adata, n_features = 5000) -> tuple:
    '''
    Function to calculate approximate kernels.
    Input:
            adata- Adata obj as created by `Create_Adata`
                Sigma can be calculated with Estimate_Sigma or a heuristic but must be positive.
            n_features- Number of random feature to use when calculating Z- used for scalability
    Output:
            adata_object with Z matrices accessible with- adata.uns['Z_train'] and adata.uns['Z_test'] respectively

    '''
    assert np.all(adata.uns['sigma'] > 0), 'Sigma must be positive'

    #Number of groupings taking from group_dict
    N_pathway = len(adata.uns['group_dict'].keys())
    D = adata.uns['D']

    # Create Arrays to store concatenated group Z.  Each group of features will have a corresponding entry in each array
    Z_train = np.zeros((len(adata.uns['train_indices']), 2 * adata.uns['D'] * N_pathway))
    Z_test = np.zeros((len(adata.uns['test_indices']), 2 * adata.uns['D'] * N_pathway))

    # Loop over each of the groups and creating Z for each
    for m, group_features in enumerate(adata.uns['group_dict'].values()):
        
        #Extract features from mth group
        num_group_features = len(group_features)

        # Sample up to n_features features- important for scalability if using large groupings
        # Will use all features if the grouping contains fewer than n_features
        group_features = adata.uns['seed_obj'].choice(np.array(list(group_features)), np.min([n_features, num_group_features]), replace = False) 

        # Create data arrays containing only features within this group
        X_train = adata[adata.uns['train_indices'],:][:, group_features].X
        X_test = adata[adata.uns['test_indices'],:][:, group_features].X

        # Perform data filtering, and transformation according to given data_type
        # Will remove low variance (< 1e5) features regardless of data_type
        # If given data_type is 'counts' (like RNA) will log scale and z-score the data
        X_train, X_test = process_data(X_train = X_train, X_test = X_test, data_type = adata.uns['data_type'], return_dense = True)

        #Extract pre-calculated sigma used for approximating kernel
        adjusted_sigma = adata.uns['sigma'][m]

        #Calculates approximate kernel according to chosen kernel function- may add more functions in the future
        #Distribution data comes from Fourier Transform of original kernel function
        if adata.uns['kernel_type'].lower() == 'gaussian':

            gamma = 1/(2*adjusted_sigma**2)
            sigma_p = 0.5*np.sqrt(2*gamma)

            W = adata.uns['seed_obj'].normal(0, sigma_p, X_train.shape[1]*D).reshape((X_train.shape[1]),D)

        elif adata.uns['kernel_type'].lower() == 'laplacian':

            gamma = 1/(2*adjusted_sigma)

            W = gamma * adata.uns['seed_obj'].standard_cauchy(X_train.shape[1]*D).reshape((X_train.shape[1],D))

        elif adata.uns['kernel_type'].lower() == 'cauchy':

            gamma = 1/(2*adjusted_sigma**2)
            b = 0.5*np.sqrt(gamma)

            W = adata.uns['seed_obj'].laplace(0, b, X_train.shape[1]*D).reshape((X_train.shape[1],D))


        train_projection = np.matmul(X_train, W)
        test_projection = np.matmul(X_test, W)
        

        #Store group Z in whole-Z object.  Preserves order to be able to extract meaningful groups
        Z_train[0:, np.arange( m * 2 * D , (m + 1) * 2 * D)] = np.sqrt(1/D)*np.hstack((np.cos(train_projection), np.sin(train_projection)))
        Z_test[0:, np.arange( m * 2 * D , (m + 1) * 2 * D)] = np.sqrt(1/D)*np.hstack((np.cos(test_projection), np.sin(test_projection)))

    adata.uns['Z_train'] = Z_train
    adata.uns['Z_test'] = Z_test


    return adata


__all__ = ['approx_kernels', 'atac_grouping', 'data_processing', 'estimate_sigma', 
           'multimodal_processing', 'optimize_alpha', 'test', 'tfidf',
           'train', 'utils']

from scmkl.approx_kernels import *
from scmkl.atac_grouping import *
from scmkl.data_processing import *
from scmkl.estimate_sigma import *
from scmkl.multimodal_processing import *
from scmkl.optimize_alpha import *
from scmkl.test import *
from scmkl.tfidf import *
from scmkl.train import *
from scmkl.utils import *


import pandas as pd
import numpy as np
import re


def _find_overlap(start1 : int, end1 : int, start2 : int, end2 : int) -> bool:
    '''
    Function to determine whether two regions on the same chromosome overlap.
    Input:
        start1 : the start position for region 1
        end1 : the end position for region 1
        start2 : the start position for region 2
        end2: the end postion for region 2
    Output:
        True if the regions overlap by 1bp
        False if the regions do not overlap
    '''
    return max(start1,start2) <= min(end1, end2)


def _get_tss(row : pd.DataFrame) -> int:
    '''
    Takes a row from a DataFrame as a DataFrame with columns 
    ['start', 'end', 'strand'] and returns the transcription start site
    depending on gene strandedness.
    Args:
        row - a pd.DataFrame from a gtf as a pd.DataFrame containing
            only one row and columns ['start', 'end', 'strand'].
    Returns:
        The transcription start site for row's respective annotation.
    '''
    if row.iloc[2] == '+':
        return row.iloc[0]
    
    elif row.iloc[2] == '-':
        return row.iloc[1]
    
    else:
        raise ValueError('Strand symbol must be "+" or "-"')
    

def _calc_range(row : pd.DataFrame, len_up : int, len_down : int) -> list:
    '''
    Returns an infered promotor region for given annotation range 
    depending on transcription start site and user-defined upstream and
    downstream adjustments.
    Args:
        row - a pd.DataFrame from a gtf as a pd.DataFrame containing
            only one row and columns ['tss', 'strand'] where 'tss' is
            the transcription start site.
        len_up - how many base pairs upstream from the transcription 
            start site the promotor range should be adjusted for.
        len_down - how many base pairs downstream from the 
            transcription start site the promotor range should be 
            adjusted for.
    Returns:
        A 2 element list where the first element is the adjusted start
        position and the second the the adjusted end position.
    '''
    if row.iloc[1] == '+':
        start = row.iloc[0] - len_up
        end = row.iloc[0] + len_down

    elif row.iloc[1] == '-':
        start = row.iloc[0] - len_down
        end = row.iloc[0] + len_up

    else:
        raise ValueError('Strand symbol must be "+" or "-"')
    
    return [start, end]


def _adjust_regions(gene_anno : pd.DataFrame, len_up : int, len_down : int
                    ) -> pd.DataFrame:
    '''
    Takes a GTF file as a pd.DataFrame and adjusts start and end 
    positions to represent promotor regions given user-defined 
    upstream and downstream adjustments relative to the transcription
    start site.
    Args:
        gene_anno - a pd.DataFrame with columns ['chr', 'start', 'end', 
            'strand', 'gene_name'] created from a GTF file.
        len_up - an int for how many base pairs upstream of the 
            transcription start site the promotor region should be 
            adjusted for.
        len_down - an int for how many base pairs downstream of the 
            transcription start site the promotor region should be 
            adjusted for.
    Returns:
        gene_anno as a pd.DataFrame where ['start', 'end'] columns 
        represent the start and end positions of inferred promotor
        regions for each annotation.
    '''
    # Subsetting DataFrame to only required data
    required_cols = ['chr', 'start', 'end', 'strand', 'gene_name']
    gene_anno = gene_anno[required_cols]

    # Adding a column that annotates transcription start site based on strand
    tss = gene_anno[['start', 'end', 'strand']].apply(_get_tss, axis = 1)
    gene_anno.insert(1, column = 'tss', value = tss)

    # Calculating start and end positions based on the tss
    adj_regions = gene_anno[['tss', 'strand']].apply(lambda row:
                                                    _calc_range(
                                                        row = row,
                                                        len_up = len_up,
                                                        len_down = len_down
                                                        ), 
                                                   axis = 1, 
                                                   result_type = 'expand'
                                                   )
    
    gene_anno.loc[:, ['start', 'end']] = adj_regions.values
    
    return gene_anno


def _create_region_dicts(gene_anno : pd.DataFrame) -> dict:
    '''
    Takes a GTF as a pd.DataFrame and returns dictionaries required for
    region comparisons between gene_annotations and assay features.
    Args:
        gene_anno - a pd.DataFrame with columns ['chr', 'start', 'end', 
            'strand', 'gene_name'] created from a GTF file.
    Returns:
        Two dictionaries. 
            peak_gene_dict is a dictionary where keys are regions and
            values are genes.
            ga_regions is a dictionary where chromosomes are keys and
            regions are values.
    '''
    peak_gene_dict = {}
    ga_regions = {}
    for i, anno in gene_anno.iterrows():
        peak_gene_dict[(anno['chr'], int(anno['start']), int(anno['end']))] = anno['gene_name']
        if anno['chr'] in ga_regions.keys():
            ga_regions[anno['chr']] = np.concatenate((ga_regions[anno['chr']], np.array([[anno['start'], anno['end']]], dtype = int)), axis = 0)
        else:
            ga_regions[anno['chr']] = np.array([[anno['start'], anno['end']]], dtype=int)

    return peak_gene_dict, ga_regions


def _create_feature_dict(feature_names : list | set | np.ndarray | pd.Series
                         ) -> dict:
    '''
    Takes an array of feature names and returns data formatted for
    feature comparisons.
    Args:
        feature_names - an iterable object of region names from 
        single-cell ATAC data matrix.
    Returns:
        A dictionary where keys are chromosomes and values are regions.
    '''
    feature_dict = {}
    feature_names = [re.split(":|-", peak) for peak in feature_names]
    for peak_set in feature_names:
        if peak_set[0] in feature_dict.keys():
            feature_dict[peak_set[0]] = np.concatenate(
                                                (feature_dict[peak_set[0]],
                                                    np.array([[peak_set[1], 
                                                                peak_set[2]]], 
                                                                dtype = int)), 
                                                                axis = 0)
        else:
            feature_dict[peak_set[0]] = np.array([[peak_set[1], 
                                                   peak_set[2]]], 
                                                   dtype = int)

    return feature_dict


def _compare_regions(feature_dict : dict, ga_regions : dict,
                     peak_gene_dict : dict, gene_sets : dict, chr_sep : str
                     ) -> dict:
    '''
    Takes features from a single-cell ATAC data matrix and regions from
    a gene annotation file to return an ATAC grouping where regions 
    from feature_dict and regions from gene annotations overlap.
    Args:
        feature_dict - a dictionary where keys are chromosomes and 
            values are regions. This data should come from a single-
            cell atac experiment.
        ga_regions - a dictionary where keys are chromosomes and values
            are regions. This data should come from a gene annotations 
            file.
        peak_gene_dict - a dictionary where keys are peaks from gene 
            annotation file and values are the gene they are associated
            with.
        gene_sets - a dictionary where keys are gene set names and 
            values are an iterable object of gene names.
        chr_sep - The character that separates the chromosome from the 
            rest of the region in the original feature array.
    Returns:
        A dictionary where keys are the names from gene_sets and values
        are a list of regions from feature_names that overlap with 
        promotor regions respective to genes in gene sets (i.e., if 
        ATAC feature in feature_names overlaps with promotor region 
        from a gene in a gene set from gene_sets, that region will be
        added to the new dictionary under the respective gene set 
        name).
    '''
    atac_grouping = {group : [] for group in gene_sets.keys()}

    for chrom in feature_dict.keys():
        if chrom not in ga_regions.keys():
            continue
        for region in feature_dict[chrom]:
            for anno in ga_regions[chrom]:
                if _find_overlap(region[0], region[1], anno[0], anno[1]):
                    gene = peak_gene_dict[(chrom, anno[0], anno[1])]
                    for group in gene_sets.keys():
                        if gene in gene_sets[group]:
                            atac_grouping[group].append(chrom + chr_sep + str(region[0]) + "-" + str(region[1]))

    return atac_grouping


def get_atac_groupings(gene_anno : pd.DataFrame, gene_sets : dict, 
                       feature_names : np.ndarray | pd.Series | list | set,
                       len_up : int = 5000, len_down : int = 5000) -> dict:
    '''
    Takes processed gene annotations in GTF format, and gene set 
    library, and feature names from a single-cell ATAC data matrix
    and creates an ATAC region set to be used with scMKL.
    Args:
        gene_anno - gene annotations in GTF format as a pd.DataFrame 
            with columns ['chr', 'start', 'end', 'strand', 'gene_name']
        gene_sets - a dictionary with gene_set names as keys and an 
            iterable object of gene_names as values.
        feature_names - an iterable, one dimension object containing 
            feature names corresponding to a single_cell ATAC data 
            matrix.
    Returns:
        A dictionary where keys are the names from gene_sets and values
        are a list of regions from feature_names that overlap with 
        promotor regions respective to genes in gene sets (i.e., if 
        ATAC feature in feature_names overlaps with promotor region 
        from a gene in a gene set from gene_sets, that region will be
        added to the new dictionary under the respective gene set 
        name).
    '''
    # Getting a unique set of gene names from gene_sets
    all_genes = {gene for group in gene_sets.keys()
                 for gene in gene_sets[group]}
    
    # Filtering out NaN values
    all_genes = [gene for gene in all_genes if type(gene) != float]

    # Filtering out annotations for genes not present in gene_sets
    gene_anno = gene_anno[np.isin(gene_anno['gene_name'], all_genes)]

    # Adjusting start and end columns to represent promotor regions
    gene_anno = _adjust_regions(gene_anno = gene_anno, 
                                len_up = len_up, len_down = len_down)

    # Creating a dictionary from assay features where [chr] : (start, end)
    feature_dict = _create_feature_dict(feature_names)

    # Creating data structures from gene_anno for comparing regions
    peak_gene_dict, ga_regions = _create_region_dicts(gene_anno)

    # Capturing the separator type used in assay
    chr_sep = '-' if '-' in feature_names[0] else ':'

    atac_groupings = _compare_regions(feature_dict = feature_dict,
                                     ga_regions = ga_regions,
                                     peak_gene_dict = peak_gene_dict,
                                     gene_sets = gene_sets,
                                     chr_sep = chr_sep)
    
    return atac_groupings


import numpy as np
import celer



def train_model(adata, group_size = 1, alpha = 0.9):
    '''
    Function to fit a grouplasso model to the provided data.
    Inputs:
            Adata with Z matrices in adata.uns
            group_size- Argument describing how the features are grouped. 
                    From Celer documentation:
                    "groupsint | list of ints | list of lists of ints.
                        Partition of features used in the penalty on w. 
                            If an int is passed, groups are contiguous blocks of features, of size groups. 
                            If a list of ints is passed, groups are assumed to be contiguous, group number g being of size groups[g]. 
                            If a list of lists of ints is passed, groups[g] contains the feature indices of the group number g."
                    If 1, model will behave identically to Lasso Regression
            alpha- Group Lasso regularization coefficient. alpha is a floating point value controlling model solution sparsity. Must be a positive float.
                        The smaller the value, the more feature groups will be selected in the trained model.
    Outputs:

            adata object with trained model in uns accessible with- adata.uns['model']
                Specifics of model:
                    model- The trained Celer Group Lasso model.  Used in Find_Selected_Pathways() function for interpretability.
                        For more information about attributes and methods see the Celer documentation at https://mathurinm.github.io/celer/generated/celer.GroupLasso.html.

    '''
    assert alpha > 0, 'Alpha must be positive'

    y_train = adata.obs['labels'].iloc[adata.uns['train_indices']]
    X_train = adata.uns['Z_train']

    cell_labels = np.unique(y_train)

    # This is a regression algorithm. We need to make the labels 'continuous' for classification, but they will remain binary.
    # Casts training labels to array of -1,1
    train_labels = np.ones(y_train.shape)
    train_labels[y_train == cell_labels[1]] = -1

    # Alphamax is a calculation to regularize the effect of alpha (a sparsity parameter) across different data sets
    alphamax = np.max(np.abs(X_train.T.dot(train_labels))) / X_train.shape[0] * alpha

    # Instantiate celer Group Lasso Regression Model Object
    model = celer.GroupLasso(groups = group_size, alpha = alphamax)

    # Fit model using training data
    model.fit(X_train, train_labels.ravel())

    adata.uns['model'] = model
    return adata